# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import io
import os
from urllib.parse import urlparse
from io import BytesIO
from speakeasy.errors import NetworkEmuError


def is_empty(bio):
    if len(bio.getbuffer()) == bio.tell():
        return True
    return False


def normalize_response_path(path):
    def _get_speakeasy_root():
        return os.path.join(os.path.dirname(__file__), os.pardir)

    root_var = '$ROOT$'

    if root_var in path:
        root = _get_speakeasy_root()
        return path.replace(root_var, root)

    return path


class Socket(object):
    """
    Represents a Windows network socket
    """
    def __init__(self, fd, family, stype, protocol, flags):
        self.fd = fd
        self.family = family
        self.type = stype
        self.protocol = protocol
        self.flags = flags
        self.connected_host = ''
        self.connected_port = 0
        self.curr_packet = BytesIO(b'')
        self.packet_queue = []

    def get_fd(self):
        return self.fd

    def get_type(self):
        return self.type

    def set_connection_info(self, host, port):
        self.connected_host = host
        self.connected_port = port

    def get_connection_info(self):
        return (self.connected_host, self.connected_port)

    def fill_recv_queue(self, responses):

        for resp in responses:
            mode = resp.get('mode', '')
            if mode.lower() == 'default':
                default_resp_path = resp.get('path')
                if default_resp_path:
                    default_resp_path = normalize_response_path(default_resp_path)
                    with open(default_resp_path, 'rb') as f:
                        self.curr_packet = BytesIO(f.read())

    def get_recv_data(self, size, peek=False):

        data = self.curr_packet.read(size)
        if not peek:
            return data
        elif peek:
            self.curr_packet.seek(-size, os.SEEK_CUR)
        return data


class WSKSocket(Socket):
    """
    Represents a WSK socket used in kernel mode applications
    """
    def __init__(self, fd, family, stype, protocol, flags):
        super(WSKSocket, self).__init__(self, fd, family, stype,
                                        protocol, flags)


class WininetComponent(object):
    """
    Base class used for WinInet connections
    """

    curr_handle = 0x20
    config = None

    def __init__(self):
        super(WininetComponent, self).__init__()
        self.handle = self.new_handle()

    def new_handle(self):
        tmp = WininetComponent.curr_handle
        WininetComponent.curr_handle += 4
        return tmp

    def get_handle(self):
        return self.handle


class WininetRequest(WininetComponent):
    """
    WinInet request object
    """
    def __init__(self, session, verb, objname, ver, ref, accepts, flags, ctx):
        super(WininetRequest, self).__init__()

        # The WiniNet APIs default to a HTTP "GET" if no verb is specified
        if not verb:
            self.verb = 'get'
        else:
            self.verb = verb.lower()

        self.objname = objname
        if not self.objname:
            self.objname = ''
        self.objname = urlparse(self.objname)

        self.session = session

        if not ver:
            ver = 'HTTP/1.1'
        self.ver = ver
        self.referrer = ref
        self.accept_types = accepts
        self.flags = flags
        self.ctx = ctx
        self.response = None

    def get_session(self):
        return self.session

    def get_server(self):
        return self.get_session().server

    def get_port(self):
        return self.get_session().port

    def get_instance(self):
        sess = self.get_session()
        return sess.get_instance()

    def is_secure(self):
        if 'INTERNET_FLAG_SECURE' in self.flags:
            return True
        return False

    def format_http_request(self, headers=None):
        request_string = ''
        action = '%s %s %s\n' % (self.verb.upper(), self.objname.path,
                                 self.ver.upper())

        request_string += action
        if headers:
            request_string += headers

        inst = self.get_instance()
        sess = self.get_session()

        host = sess.server
        request_string += 'Host: %s\n' % (host)

        ua = inst.get_user_agent()
        if ua:
            request_string += 'User-Agent: %s\n' % (ua)

        if 'INTERNET_FLAG_KEEP_CONNECTION' in self.flags:
            request_string += 'Connection: Keep-Alive\n'
        else:
            request_string += 'Connection: Close\n'

        if 'INTERNET_FLAG_DONT_CACHE' in self.flags:
            request_string += 'Cache-Control: no-cache\n'

        return request_string

    def get_response_size(self):
        resp = self.get_response()
        off = resp.tell()
        size = len(resp.read())
        resp.seek(off, io.SEEK_SET)
        return size

    def get_response(self):
        """
        Check the configuration file so see if there is a
        handler for the current WinInet request
        """

        cfg = WininetComponent.config

        if self.response:
            return self.response

        http = cfg.get('http')
        if not http:
            raise NetworkEmuError('No HTTP configuration supplied')
        resps = http.get('responses')
        if not resps:
            raise NetworkEmuError('No HTTP responses supplied')

        self.response = None
        for res in resps:
            verb = res.get('verb', '')
            if verb.lower() == self.verb:

                resp_files = res.get('files', [])
                if resp_files:
                    for file in resp_files:
                        mode = file.get('mode', '')
                        if mode.lower() == 'by_ext':
                            ext = file.get('ext', '')
                            fn, obj_ext = os.path.splitext(self.objname.path)

                            if (ext.lower().strip('.') ==
                               obj_ext.lower().strip('.')):
                                path = file.get('path')
                                path = normalize_response_path(path)

                                with open(path, 'rb') as f:
                                    self.response = BytesIO(f.read())
                        elif mode.lower() == 'default':

                            default_resp_path = file.get('path')
                            default_resp_path = normalize_response_path(default_resp_path)

                    if not self.response and default_resp_path:
                        default_resp_path = normalize_response_path(default_resp_path)
                        with open(default_resp_path, 'rb') as f:
                            self.response = BytesIO(f.read())

        return self.response

    def get_object_path(self):
        return self.objname


class WininetSession(WininetComponent):
    def __init__(self, instance, server, port, user,
                 password, service, flags, ctx):
        super(WininetSession, self).__init__()
        self.server = server
        self.port = port
        self.user = user
        self.password = password
        self.service = service
        self.flags = flags
        self.ctx = ctx
        self.requests = {}

        self.instance = instance

    def get_instance(self):
        return self.instance

    def get_flags(self):
        return self.flags

    def new_request(self, verb, objname, ver, ref, accepts, flags, ctx):
        req = WininetRequest(self, verb, objname, ver, ref,
                             accepts, flags, ctx)
        hdl = req.get_handle()
        self.requests.update({hdl: req})
        return req


class WininetInstance(WininetComponent):
    def __init__(self, user_agent, access, proxy, bypass, flags):
        super(WininetInstance, self).__init__()
        self.user_agent = user_agent
        self.access = access
        self.proxy = proxy
        self.bypass = bypass
        self.flags = flags
        self.sessions = {}

    def get_session(self, sess_handle):
        self.sessions.get(sess_handle)

    def add_session(self, handle, session):
        self.sessions.update({handle: session})

    def new_session(self, server, port, user, password, service, flags, ctx):
        sess = WininetSession(self, server, port, user,
                              password, service, flags, ctx)
        hdl = sess.get_handle()
        self.sessions.update({hdl: sess})
        return sess

    def get_user_agent(self):
        return self.user_agent


class NetworkManager(object):
    """
    Class that manages network connections during emulation
    """
    def __init__(self, config):
        super(NetworkManager, self).__init__()
        self.sockets = {}
        self.wininets = {}
        self.curr_fd = 4
        self.curr_handle = 0x20
        self.config = config
        self.dns = {}

        WininetComponent.config = config
        self.dns = self.config.get('dns')

    def new_socket(self, family, stype, protocol, flags):

        fd = self.curr_fd

        sock = Socket(fd, family, stype, protocol, flags)
        self.curr_fd += 4

        if self.config:
            winsock = self.config.get('winsock')
            if winsock:
                responses = winsock.get('responses')
                if responses:
                    sock.fill_recv_queue(responses)

        self.sockets.update({fd: sock})
        return sock

    def name_lookup(self, domain):

        if not self.dns:
            return None

        names = self.dns.get('names')

        # Do we have an IP for this name?
        if domain.lower() not in names.keys():
            # use the default IP (if any)
            return names.get('default')

        return names.get(domain)

    def get_dns_txt(self, domain):
        """
        Return a configured DNS TXT record (if any)
        """
        def _read_txt_data(txt):
            path = txt.get('path')
            if path:
                path = normalize_response_path(path)
                with open(path, 'rb') as f:
                    return f.read()

        if not self.dns:
            return None

        txts = self.dns.get('txt', [])
        txt = [t for t in txts if t.get('name', '') == domain]
        if txt:
            return _read_txt_data(txt[0])
        txt = [t for t in txts if t.get('name', '') == 'default']
        if txt:
            return _read_txt_data(txt[0])

    def ip_lookup(self, ip):
        for item in self.dns:
            if item['response'] == ip:
                return item['query']
        return None

    def new_wininet_inst(self, user_agent, access, proxy, bypass, flags):
        wini = WininetInstance(user_agent, access, proxy, bypass, flags)

        self.wininets.update({wini.get_handle(): wini})
        return wini

    def get_wininet_object(self, handle):

        for hinst, inst in self.wininets.items():
            if hinst == handle:
                return inst
            for hsess, sess in inst.sessions.items():
                if hsess == handle:
                    return sess
                for hreq, req in sess.requests.items():
                    if hreq == handle:
                        return req

    def close_wininet_object(self, handle):
        if self.wininets.get(handle):
            self.wininets.pop(handle)

    def get_socket(self, fd):
        return self.sockets.get(fd)

    def close_socket(self, fd):
        self.sockets.pop(fd)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
import speakeasy.winenv.defs.registry.reg as regdefs
import speakeasy.winenv.defs.windows.windows as windefs

import speakeasy.winenv.defs.windows.kernel32 as k32
import speakeasy.winenv.defs.windows.advapi32 as adv32
import speakeasy.windows.objman as objman

from .. import api
from Crypto.Cipher import ARC4
import hashlib

SERVICE_STATUS_HANDLE_BASE = 0x1000


class AdvApi32(api.ApiHandler):
    """
    Implements exported functions from advapi32.dll
    """

    name = 'advapi32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):
        super(AdvApi32, self).__init__(emu)
        self.funcs = {}
        self.data = {}
        self.hash_objects = {}
        self.key_objects = {}
        self.k32types = k32
        self.win = adv32
        self.curr_rand = 0
        self.curr_handle = 0x2800
        self.service_status_handle = SERVICE_STATUS_HANDLE_BASE

        self.rc4 = None

        super(AdvApi32, self).__get_hook_attrs__(self)

    def get_handle(self):
        self.curr_handle += 4
        return self.curr_handle

    @apihook('RegOpenKey', argc=3, conv=_arch.CALL_CONV_STDCALL)
    def RegOpenKey(self, emu, argv, ctx={}):
        '''
        LSTATUS RegOpenKeyA(
          HKEY   hKey,
          LPCSTR lpSubKey,
          PHKEY  phkResult
        );
        '''

        hKey, lpSubKey, phkResult = argv
        rv = windefs.ERROR_SUCCESS
        hnd = 0

        hkey_name = regdefs.get_hkey_type(hKey)
        if hkey_name:
            argv[0] = hkey_name
            if not hnd and not lpSubKey:
                hnd = hKey
        else:
            key_obj = emu.regman.get_key_from_handle(hKey)
            if not key_obj:
                return windefs.ERROR_PATH_NOT_FOUND
            hkey_name = key_obj.path

        cw = self.get_char_width(ctx)
        if lpSubKey:
            lpSubKey = self.read_mem_string(lpSubKey, cw)
            argv[1] = lpSubKey

            if hkey_name and lpSubKey:
                if not lpSubKey.startswith('\\'):
                    lpSubKey = '\\' + lpSubKey
                lpSubKey = hkey_name + lpSubKey

            hnd = self.reg_open_key(lpSubKey, create=False)
            if not hnd:
                rv = windefs.ERROR_PATH_NOT_FOUND

            self.log_registry_access(lpSubKey, 'open_key', handle=hnd)

        if phkResult and hnd:
            self.mem_write(phkResult, hnd.to_bytes(self.get_ptr_size(), 'little'))

        return rv

    @apihook('RegOpenKeyEx', argc=5, conv=_arch.CALL_CONV_STDCALL)
    def RegOpenKeyEx(self, emu, argv, ctx={}):
        '''
        LSTATUS RegOpenKeyEx(
          HKEY   hKey,
          LPTSTR lpSubKey,
          DWORD  ulOptions,
          REGSAM samDesired,
          PHKEY  phkResult
        );
        '''

        hKey, lpSubKey, ulOptions, samDesired, phkResult = argv
        rv = windefs.ERROR_SUCCESS

        hnd = 0

        hkey_name = regdefs.get_hkey_type(hKey)
        if hkey_name:
            argv[0] = hkey_name
            if not hnd and not lpSubKey:
                hnd = hKey

        cw = self.get_char_width(ctx)
        if lpSubKey:
            lpSubKey = self.read_mem_string(lpSubKey, cw)
            argv[1] = lpSubKey

            if hkey_name and lpSubKey:
                if not lpSubKey.startswith('\\'):
                    lpSubKey = '\\' + lpSubKey
                lpSubKey = hkey_name + lpSubKey

            hnd = self.reg_open_key(lpSubKey, create=False)
            if not hnd:
                rv = windefs.ERROR_PATH_NOT_FOUND

            self.log_registry_access(lpSubKey, 'open_key', handle=hnd)

        if phkResult and hnd:
            self.mem_write(phkResult, hnd.to_bytes(self.get_ptr_size(), 'little'))

        return rv

    @apihook('RegQueryValueEx', argc=6, conv=_arch.CALL_CONV_STDCALL)
    def RegQueryValueEx(self, emu, argv, ctx={}):
        '''
        LSTATUS RegQueryValueEx(
          HKEY    hKey,
          LPTSTR  lpValueName,
          LPDWORD lpReserved,
          LPDWORD lpType,
          LPBYTE  lpData,
          LPDWORD lpcbData
        );
        '''

        hKey, lpValueName, lpReserved, lpType, lpData, lpcbData = argv
        rv = windefs.ERROR_SUCCESS

        cw = self.get_char_width(ctx)
        if lpValueName:
            lpValueName = self.read_mem_string(lpValueName, cw)
            argv[1] = lpValueName

        type_name = regdefs.get_value_type(lpType)
        if type_name:
            argv[3] = type_name

        length = 0
        if lpcbData:
            length = self.mem_read(lpcbData, 4)
            length = int.from_bytes(length, 'little')
            argv[5] = length

        key = self.reg_get_key(hKey)
        if key:
            val = key.get_value(lpValueName)
            if val:
                output = b''
                typ = val.get_type()
                data = val.get_data()
                if typ == 'REG_SZ':
                    output = data.encode('utf-8')

                if lpcbData:
                    self.mem_write(lpcbData, len(output).to_bytes(4, 'little'))

                if len(output) > length:
                    rv = windefs.ERROR_INSUFFICIENT_BUFFER
                else:
                    if lpData:
                        self.mem_write(lpData, output)

            # For now, return an empty buffer
            else:
                output = b'\x00' * length
                if lpData:
                    try:
                        self.mem_write(lpData, output)
                    except Exception:
                        return windefs.ERROR_INVALID_PARAMETER
                if lpcbData:
                    self.mem_write(lpcbData, len(output).to_bytes(4, 'little'))
                rv = windefs.ERROR_SUCCESS

            kp = key.get_path()
            self.log_registry_access(kp, 'read_value', value_name=lpValueName, size=length,
                                     buffer=lpData)

        return rv

    @apihook('RegCloseKey', argc=1, conv=_arch.CALL_CONV_STDCALL)
    def RegCloseKey(self, emu, argv, ctx={}):
        '''
        LSTATUS RegCloseKey(
          HKEY hKey
        );
        '''

        hKey, = argv
        rv = windefs.ERROR_SUCCESS

        key = self.reg_get_key(hKey)
        if not key:
            rv = windefs.ERROR_INVALID_HANDLE

        return rv

    @apihook('RegEnumKey', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def RegEnumKey(self, emu, argv, ctx={}):
        '''
        LSTATUS RegEnumKey(
          HKEY  hKey,
          DWORD dwIndex,
          LPTSTR lpName,
          DWORD cchName
        );
        '''

        hKey, dwIndex, lpName, cchName = argv

        _argv = argv + [0, 0, 0, 0]
        rv = self.RegEnumKeyEx(emu, _argv, ctx)
        argv[:] = _argv[: 4]

        return rv

    @apihook('RegEnumKeyEx', argc=8, conv=_arch.CALL_CONV_STDCALL)
    def RegEnumKeyEx(self, emu, argv, ctx={}):
        '''
        LSTATUS RegEnumKeyEx(
            HKEY      hKey,
            DWORD     dwIndex,
            LPSTR     lpName,
            LPDWORD   lpcchName,
            LPDWORD   lpReserved,
            LPSTR     lpClass,
            LPDWORD   lpcchClass,
            PFILETIME lpftLastWriteTime
        );
        '''

        hKey, dwIndex, lpName, cchName, res, pcls, cchcls, last_write = argv

        cw = self.get_char_width(ctx)
        rv = windefs.ERROR_INVALID_HANDLE
        if hKey:
            key = self.reg_get_key(hKey)
            argv[0] = key.get_path()
            if not key:
                rv = windefs.ERROR_INVALID_HANDLE
            else:
                subkeys = self.reg_get_subkeys(key)
                if (dwIndex + 1) > len(subkeys):
                    rv = windefs.ERROR_NO_MORE_ITEMS
                else:
                    if lpName:
                        sk = subkeys[dwIndex]
                        name = sk.get_path()
                        if cw == 2:
                            name = name.encode('utf-16le')
                        else:
                            name = name.encode('utf-8')
                        self.mem_write(lpName, name)
                        rv = windefs.ERROR_SUCCESS
            self.log_registry_access(key.get_path(), "list_subkeys")
        return rv

    @apihook('RegCreateKey', argc=3)
    def RegCreateKey(self, emu, argv, ctx={}):
        """
        LSTATUS RegCreateKey(
            HKEY    hKey,
            LPCWSTR lpSubKey,
            PHKEY   phkResult
        );
        """
        hkey, lpSubKey, phkResult = argv
        rv = windefs.ERROR_INVALID_HANDLE
        if hkey:
            key = self.reg_get_key(hkey)
            argv[0] = key.get_path()
            if not key:
                rv = windefs.ERROR_INVALID_HANDLE
            else:
                cw = self.get_char_width(ctx)
                if lpSubKey:
                    lpSubKey = self.read_mem_string(lpSubKey, cw)
                    argv[1] = lpSubKey
                    sub_key_path = key.get_path() + '\\' + lpSubKey
                    self.emu.reg_create_key(sub_key_path)
                    self.log_registry_access(sub_key_path, "create_key")
                else:
                    hkey = (hkey).to_bytes(self.get_ptr_size(), 'little')
                    self.mem_write(phkResult, hkey)
                    rv = windefs.ERROR_SUCCESS
        return rv

    @apihook('RegQueryInfoKey', argc=12, conv=_arch.CALL_CONV_STDCALL)
    def RegQueryInfoKey(self, emu, argv, ctx={}):
        # TODO: stub
        '''
        LSTATUS RegQueryInfoKeyA(
          HKEY      hKey,
          LPSTR     lpClass,
          LPDWORD   lpcchClass,
          LPDWORD   lpReserved,
          LPDWORD   lpcSubKeys,
          LPDWORD   lpcbMaxSubKeyLen,
          LPDWORD   lpcbMaxClassLen,
          LPDWORD   lpcValues,
          LPDWORD   lpcbMaxValueNameLen,
          LPDWORD   lpcbMaxValueLen,
          LPDWORD   lpcbSecurityDescriptor,
          PFILETIME lpftLastWriteTime
        );
        '''

        hKey, lpClass, lpcchClass, _, subkeys, max_subkey_len, max_class_len, \
            values, max_value_name_len, max_value_len, sec_desc, last_write = argv

        rv = windefs.ERROR_SUCCESS

        hkey_name = regdefs.get_hkey_type(hKey)
        if hkey_name:
            argv[0] = hkey_name

        key = self.reg_get_key(hKey)
        if not key:
            rv = windefs.ERROR_INVALID_HANDLE

        return rv

    @apihook('OpenProcessToken', argc=3, conv=_arch.CALL_CONV_STDCALL)
    def OpenProcessToken(self, emu, argv, ctx={}):
        '''
        BOOL OpenProcessToken(
          HANDLE  ProcessHandle,
          DWORD   DesiredAccess,
          PHANDLE pTokenHandle
        );
        '''

        hProcess, DesiredAccess, pTokenHandle = argv
        rv = 0

        if hProcess == self.get_max_int():
            obj = emu.get_current_process()
        else:
            obj = self.get_object_from_handle(hProcess)

        if obj:
            token = obj.get_token()
            hToken = token.get_handle()

            if pTokenHandle:
                hnd = (hToken).to_bytes(self.get_ptr_size(), 'little')
                self.mem_write(pTokenHandle, hnd)
                rv = 1
                emu.set_last_error(windefs.ERROR_SUCCESS)
            else:
                emu.set_last_error(windefs.ERROR_INVALID_PARAMETER)

        return rv

    @apihook('OpenThreadToken', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def OpenThreadToken(self, emu, argv, ctx={}):
        '''
        BOOL OpenThreadToken(
            HANDLE  ThreadHandle,
            DWORD   DesiredAccess,
            BOOL    OpenAsSelf,
            PHANDLE TokenHandle
        );
        '''

        ThreadHandle, DesiredAccess, OpenAsSelf, pTokenHandle = argv
        rv = 0

        if ThreadHandle == self.get_max_int():
            obj = emu.get_current_thread()
        else:
            obj = self.get_object_from_handle(ThreadHandle)

        if obj:
            token = obj.get_token()
            hToken = token.get_handle()

            if pTokenHandle:
                hnd = (hToken).to_bytes(self.get_ptr_size(), 'little')
                self.mem_write(pTokenHandle, hnd)
                rv = 1
                emu.set_last_error(windefs.ERROR_SUCCESS)
            else:
                emu.set_last_error(windefs.ERROR_INVALID_PARAMETER)

        return rv

    @apihook('DuplicateTokenEx', argc=6, conv=_arch.CALL_CONV_STDCALL)
    def DuplicateTokenEx(self, emu, argv, ctx={}):
        '''
        BOOL DuplicateTokenEx(
          HANDLE                       hExistingToken,
          DWORD                        dwDesiredAccess,
          LPSECURITY_ATTRIBUTES        lpTokenAttributes,
          SECURITY_IMPERSONATION_LEVEL ImpersonationLevel,
          TOKEN_TYPE                   TokenType,
          PHANDLE                      phNewToken
        );
        '''

        (hExistingToken, access, token_attrs, imp_level, toktype,
         phNewToken) = argv
        rv = 0

        obj = self.get_object_from_handle(hExistingToken)

        if obj:

            new_token = emu.new_object(objman.Token)
            hnd_new_token = new_token.get_handle()

            if phNewToken:
                hnd = (hnd_new_token).to_bytes(self.get_ptr_size(), 'little')
                self.mem_write(phNewToken, hnd)
                rv = 1
                emu.set_last_error(windefs.ERROR_SUCCESS)
            else:
                emu.set_last_error(windefs.ERROR_INVALID_PARAMETER)

        return rv

    @apihook('SetTokenInformation', argc=4, conv=_arch.CALL_CONV_STDCALL)
    def SetTokenInformation(self, emu, argv, ctx={}):
        '''
        BOOL SetTokenInformation(
          HANDLE                  TokenHandle,
          TOKEN_INFORMATION_CLASS TokenInformationClass,
          LPVOID                  TokenInformation,
          DWORD                   TokenInformationLength
        );
        '''

        handle, info_class, info, info_len = argv

        rv = 1

        return rv

    @apihook('StartServiceCtrlDispatcher', argc=1)
    def StartServiceCtrlDispatcher(self, emu, argv, ctx={}):
        '''
        BOOL StartServiceCtrlDispatcher(
          const SERVICE_TABLE_ENTRY *lpServiceStartTable
        );
        '''
        lpServiceStartTable, = argv

        cw = self.get_char_width(ctx)

        ste = self.win.SERVICE_TABLE_ENTRY(emu.get_ptr_size())
        entry = self.mem_cast(ste, lpServiceStartTable)

        argv[0] = "lpServiceStartTable=["

        while (entry.lpServiceName != windefs.NULL or
                entry.lpServiceProc != windefs.NULL):
            # Get the service name
            if entry.lpServiceName != windefs.NULL:
                name = self.read_mem_string(entry.lpServiceName, cw) # noqa
                argv[0] += " {{ lpServiceName={}".format(name)
            else:
                argv[0] += " { lpServiceName=NULL"
            # Get the ServiceMain function
            if entry.lpServiceProc != windefs.NULL:
                service_main = entry.lpServiceProc
                argv[0] += ", lpServiceProc={} }} ".format(hex(service_main))
                handle, obj = self.create_thread(service_main, windefs.NULL,
                                                 emu.get_current_process())
            else:
                argv[0] += ", lpServiceProc=NULL } "
            # next entry
            lpServiceStartTable += self.sizeof(ste)
            ste = self.win.SERVICE_TABLE_ENTRY(emu.get_ptr_size())
            entry = self.mem_cast(ste, lpServiceStartTable)

        argv[0] += "]"

        rv = True
        emu.set_last_error(windefs.ERROR_SUCCESS)

        return rv

    @apihook('RegisterServiceCtrlHandler', argc=2)
    def RegisterServiceCtrlHandler(self, emu, argv, ctx={}):
        '''
        SERVICE_STATUS_HANDLE RegisterServiceCtrlHandlerA(
            LPCSTR             lpServiceName,
            LPHANDLER_FUNCTION lpHandlerProc
            );
        '''

        lpServiceName, lpHandlerProc = argv

        # dummy SERVICE_STATUS_HANDLE
        self.service_status_handle += 1

        emu.set_last_error(windefs.ERROR_SUCCESS)

        return self.service_status_handle

    @apihook('SetServiceStatus', argc=2)
    def SetServiceStatus(self, emu, argv, ctx={}):
        '''
        BOOL SetServiceStatus(
            SERVICE_STATUS_HANDLE hServiceStatus,
            LPSERVICE_STATUS      lpServiceStatus
            );
        '''

        hServiceStatus, lpServiceStatus = argv

        emu.set_last_error(windefs.ERROR_SUCCESS)

        return 0x1

    @apihook('RevertToSelf', argc=0)
    def RevertToSelf(self, emu, argv, ctx={}):
        '''
        BOOL RevertToSelf();
        '''
        return 1

    @apihook('ImpersonateLoggedOnUser', argc=1)
    def ImpersonateLoggedOnUser(self, emu, argv, ctx={}):
        '''
        BOOL ImpersonateLoggedOnUser(
        HANDLE hToken
        );
        '''
        return 1

    @apihook('OpenSCManager', argc=3)
    def OpenSCManager(self, emu, argv, ctx={}):
        '''
        SC_HANDLE OpenSCManager(
          LPCSTR lpMachineName,
          LPCSTR lpDatabaseName,
          DWORD  dwDesiredAccess
        );
        '''
        lpMachineName, lpDatabaseName, dwDesiredAccess = argv

        hScm = self.mem_alloc(size=8)
        emu.set_last_error(windefs.ERROR_SUCCESS)

        return hScm

    @apihook('CreateService', argc=13)
    def CreateService(self, emu, argv, ctx={}):
        '''
        SC_HANDLE CreateServiceA(
          SC_HANDLE hSCManager,
          LPCSTR    lpServiceName,
          LPCSTR    lpDisplayName,
          DWORD     dwDesiredAccess,
          DWORD     dwServiceType,
          DWORD     dwStartType,
          DWORD     dwErrorControl,
          LPCSTR    lpBinaryPathName,
          LPCSTR    lpLoadOrderGroup,
          LPDWORD   lpdwTagId,
          LPCSTR    lpDependencies,
          LPCSTR    lpServiceStartName,
          LPCSTR    lpPassword
        );
        '''
        (hScm, svc_name, disp_name, access,
         svc_type, start_type, error_ctrl, bin_path,
         load_group, tag_id, deps, svc_start_name,
         password) = argv

        cw = self.get_char_width(ctx)

        if svc_name:
            _sname = self.read_mem_string(svc_name, cw)
            argv[1] = _sname
        if disp_name:
            _dname = self.read_mem_string(disp_name, cw)
            argv[2] = _dname
        if bin_path:
            _bpname = self.read_mem_string(bin_path, cw)
            argv[7] = _bpname

        hSvc = self.mem_alloc(size=8)
        emu.set_last_error(windefs.ERROR_SUCCESS)

        return hSvc

    @apihook('StartService', argc=3)
    def StartService(self, emu, argv, ctx={}):
        '''
        BOOL StartService(
          SC_HANDLE hService,
          DWORD     dwNumServiceArgs,
          LPCSTR    *lpServiceArgVectors
        );
        '''
        hService, dwNumServiceArgs, lpServiceArgVectors = argv

        rv = 1

        emu.set_last_error(windefs.ERROR_SUCCESS)

        return rv

    @apihook('CloseServiceHandle', argc=1)
    def CloseServiceHandle(self, emu, argv, ctx={}):
        '''
        BOOL CloseServiceHandle(
          SC_HANDLE hSCObject
        );
        '''
        CloseServiceHandle, = argv

        self.mem_free(CloseServiceHandle)

        rv = 1

        emu.set_last_error(windefs.ERROR_SUCCESS)

        return rv

    @apihook('ChangeServiceConfig2', argc=3)
    def ChangeServiceConfig2(self, emu, argv, ctx={}):
        '''
        BOOL ChangeServiceConfig2(
          SC_HANDLE hService,
          DWORD     dwInfoLevel,
          LPVOID    lpInfo
        );
        '''
        hService, dwInfoLevel, lpInfo = argv

        rv = 1

        emu.set_last_error(windefs.ERROR_SUCCESS)

        return rv

    @apihook('SystemFunction036', argc=2)
    def RtlGenRandom(self, emu, argv, ctx={}):
        '''
        BOOLEAN RtlGenRandom(
            PVOID RandomBuffer,
            ULONG RandomBufferLength
        );
        '''
        RandomBuffer, RandomBufferLength = argv

        rv = False
        if RandomBuffer and RandomBufferLength:
            buf = bytes([i for i in range(RandomBufferLength)])
            self.mem_write(RandomBuffer, buf)
            rv = True

        return rv

    @apihook('CryptAcquireContext', argc=5)
    def CryptAcquireContext(self, emu, argv, ctx={}):
        '''
        BOOL CryptAcquireContext(
            HCRYPTPROV *phProv,
            LPCSTR     szContainer,
            LPCSTR     szProvider,
            DWORD      dwProvType,
            DWORD      dwFlags
        );
        '''
        phProv, szContainer, szProvider, dwProvType, dwFlags = argv
        cont_str, prov_str = '', ''
        cw = self.get_char_width(ctx)
        rv = False

        if szContainer:
            cont_str = self.read_mem_string(szContainer, cw)
            argv[1] = cont_str
        if szProvider:
            prov_str = self.read_mem_string(szProvider, cw)
            argv[2] = prov_str

        cm = emu.get_crypt_manager()
        hnd = cm.crypt_open(cname=cont_str, pname=prov_str, ptype=dwProvType, flags=dwFlags)

        if hnd and phProv:
            self.mem_write(phProv, hnd.to_bytes(emu.get_ptr_size(), 'little'))
            rv = True
            emu.set_last_error(windefs.ERROR_SUCCESS)

        return rv

    @apihook('CryptGenRandom', argc=3)
    def CryptGenRandom(self, emu, argv, ctx={}):
        '''
        BOOL CryptGenRandom(
            HCRYPTPROV hProv,
            DWORD      dwLen,
            BYTE       *pbBuffer
        );
        '''
        hProv, dwLen, pbBuffer = argv
        rv = False

        if pbBuffer:
            out = b'A' * dwLen
            self.mem_write(pbBuffer, out)
            rv = True

        return rv

    @apihook('AllocateAndInitializeSid', argc=11)
    def AllocateAndInitializeSid(self, emu, argv, ctx={}):
        '''
        BOOL AllocateAndInitializeSid(
            PSID_IDENTIFIER_AUTHORITY pIdentifierAuthority,
            BYTE                      nSubAuthorityCount,
            DWORD                     nSubAuthority0,
            DWORD                     nSubAuthority1,
            DWORD                     nSubAuthority2,
            DWORD                     nSubAuthority3,
            DWORD                     nSubAuthority4,
            DWORD                     nSubAuthority5,
            DWORD                     nSubAuthority6,
            DWORD                     nSubAuthority7,
            PSID                      *pSid
        );
        '''
        auth, count, sa0, sa1, sa2, sa3, sa4, sa5, sa6, sa7, pSid = argv
        rv = False

        if pSid:
            sid = self.mem_alloc(0x100, tag='api.struct.SID')
            self.mem_write(pSid, sid.to_bytes(emu.get_ptr_size(), 'little'))
            rv = True

        return rv

    @apihook('CheckTokenMembership', argc=3)
    def CheckTokenMembership(self, emu, argv, ctx={}):
        '''
        BOOL CheckTokenMembership(
            HANDLE TokenHandle,
            PSID   SidToCheck,
            PBOOL  IsMember
        );
        '''
        TokenHandle, SidToCheck, IsMember = argv
        rv = False

        if IsMember:
            self.mem_write(IsMember, (1).to_bytes(4, 'little'))
            rv = True
        return rv

    @apihook('FreeSid', argc=1)
    def FreeSid(self, emu, argv, ctx={}):
        '''
        PVOID FreeSid(
            PSID pSid
        );
        '''
        pSid,  = argv
        rv = pSid

        if pSid:
            self.mem_free(pSid)
            rv = 0
        return rv

    @apihook('CryptReleaseContext', argc=2)
    def CryptReleaseContext(self, emu, argv, ctx={}):
        '''
        BOOL CryptReleaseContext(
            HCRYPTPROV hProv,
            DWORD      dwFlags
        );
        '''
        hProv, dwFlags = argv
        rv = True

        cm = emu.get_crypt_manager()
        cm.crypt_close(hProv)

        return rv

    @apihook('GetUserName', argc=2)
    def GetUserName(self, emu, argv, ctx={}):
        '''
        BOOL GetUserName(
            LPSTR   lpBuffer,
            LPDWORD pcbBuffer
        );
        '''
        lpBuffer, pcbBuffer = argv
        rv = False
        cw = self.get_char_width(ctx)

        user = emu.get_user()
        user_name = user.get('name')
        argv[0] = user_name

        if lpBuffer:
            if cw == 2:
                out = user_name.encode('utf-16le')
            elif cw == 1:
                out = user_name.encode('utf-8')
            self.mem_write(lpBuffer, out)
            rv = True
        if pcbBuffer:
            self.mem_write(pcbBuffer, (len(user_name)).to_bytes(4, 'little'))

        return rv

    @apihook('LookupPrivilegeValue', argc=3)
    def LookupPrivilegeValue(self, emu, argv, ctx={}):
        '''
        BOOL LookupPrivilegeValue(
            LPCSTR lpSystemName,
            LPCSTR lpName,
            PLUID  lpLuid
        );
        '''
        sysname, name, luid = argv
        rv = False
        cw = self.get_char_width(ctx)

        if sysname:
            sysname = self.read_mem_string(sysname, cw)
            argv[0] = sysname
        if name:
            name = self.read_mem_string(name, cw)
            argv[1] = name
            rv = True

        return rv

    @apihook('AdjustTokenPrivileges', argc=6)
    def AdjustTokenPrivileges(self, emu, argv, ctx={}):
        '''
        BOOL AdjustTokenPrivileges(
            HANDLE            TokenHandle,
            BOOL              DisableAllPrivileges,
            PTOKEN_PRIVILEGES NewState,
            DWORD             BufferLength,
            PTOKEN_PRIVILEGES PreviousState,
            PDWORD            ReturnLength
        );
        '''
        rv = True

        return rv

    @apihook('GetTokenInformation', argc=5)
    def GetTokenInformation(self, emu, argv, ctx={}):
        '''
        BOOL GetTokenInformation(
            HANDLE                  TokenHandle,
            TOKEN_INFORMATION_CLASS TokenInformationClass,
            LPVOID                  TokenInformation,
            DWORD                   TokenInformationLength,
            PDWORD                  ReturnLength
        );
        '''
        hnd, info_class, info, info_len, ret_len = argv
        rv = True

        if not info_len:
            rv = False
            emu.set_last_error(windefs.ERROR_INSUFFICIENT_BUFFER)

        if info_class == 20 and info and emu.get_user().get('is_admin', True):
            self.mem_write(info, (1).to_bytes(4, 'little'))

        if ret_len:
            self.mem_write(ret_len, (4).to_bytes(4, 'little'))

        return rv

    @apihook('EqualSid', argc=2)
    def EqualSid(self, emu, argv, ctx={}):
        '''
        BOOL EqualSid(
            PSID pSid1,
            PSID pSid2
        );
        '''
        sid1, sid2 = argv
        rv = False

        if sid1 and sid2:
            s1 = self.mem_read(sid1, 10)
            s2 = self.mem_read(sid2, 10)
            if s1 == s2:
                rv = True

        return rv

    @apihook('GetSidSubAuthorityCount', argc=1)
    def GetSidSubAuthorityCount(self, emu, argv, ctx={}):
        '''
        PUCHAR GetSidSubAuthorityCount(
            PSID pSid
        );
        '''
        sid, = argv
        rv = 0

        if sid:
            rv = sid + 1

        return rv

    @apihook('LookupAccountSid', argc=7)
    def LookupAccountSid(self, emu, argv, ctx={}):
        '''
        BOOL LookupAccountSid(
            LPCSTR        lpSystemName,
            PSID          Sid,
            LPSTR         Name,
            LPDWORD       cchName,
            LPSTR         ReferencedDomainName,
            LPDWORD       cchReferencedDomainName,
            PSID_NAME_USE peUse
        );
        '''
        sysname, sid, name, cchname, domname, cchdomname, peuse = argv
        rv = False

        cw = self.get_char_width(ctx)

        if not cchname or not cchdomname:
            return rv

        name_size = self.mem_read(cchname, 4)
        name_size = int.from_bytes(name_size, 'little')

        dom_size = self.mem_read(cchdomname, 4)
        dom_size = int.from_bytes(dom_size, 'little')

        self.write_mem_string('myuser', name, cw)
        self.write_mem_string('mydomain', domname, cw)
        rv = True

        if sysname:
            sn = self.read_mem_string(sysname, cw)
            argv[0] = sn

        return rv

    @apihook('CreateProcessAsUser', argc=11, conv=_arch.CALL_CONV_STDCALL)
    def CreateProcessAsUser(self, emu, argv, ctx={}):
        '''
        BOOL CreateProcessAsUser(
          HANDLE                hToken,
          LPCSTR                lpApplicationName,
          LPSTR                 lpCommandLine,
          LPSECURITY_ATTRIBUTES lpProcessAttributes,
          LPSECURITY_ATTRIBUTES lpThreadAttributes,
          BOOL                  bInheritHandles,
          DWORD                 dwCreationFlags,
          LPVOID                lpEnvironment,
          LPCSTR                lpCurrentDirectory,
          LPSTARTUPINFOA        lpStartupInfo,
          LPPROCESS_INFORMATION lpProcessInformation
        );
        '''
        token, app, cmd, pa, ta, inherit, flags, env, cd, si, ppi = argv

        cw = self.get_char_width(ctx)
        cmdstr = ''
        appstr = ''
        if app:
            appstr = self.read_mem_string(app, cw)
            argv[1] = appstr
        if cmd:
            cmdstr = self.read_mem_string(cmd, cw)
            if not appstr:
                appstr = cmdstr.split(' ')[0]
            argv[2] = cmdstr

        proc = emu.create_process(path=appstr, cmdline=cmdstr)
        proc_hnd = self.get_object_handle(proc)

        thread = proc.threads[0]
        thread_hnd = self.get_object_handle(thread)

        _pi = self.k32types.PROCESS_INFORMATION(emu.get_ptr_size())
        data = self.mem_cast(_pi, ppi)
        _pi.hProcess = proc_hnd
        _pi.hThread = thread_hnd
        _pi.dwProcessId = proc.pid
        _pi.dwThreadId = thread.tid

        self.mem_write(ppi, self.get_bytes(data))

        rv = 1

        self.log_process_event(proc, 'create')
        return rv

    @apihook('CryptCreateHash', argc=5)
    def CryptCreateHash(self, emu, argv, ctx={}):
        '''
        BOOL CryptCreateHash(
          HCRYPTPROV hProv,
          ALG_ID     Algid,
          HCRYPTKEY  hKey,
          DWORD      dwFlags,
          HCRYPTHASH *phHash
        );
        '''

        hash_algs = {
            0x00008004: ('CALG_SHA1', hashlib.sha1),
            0x0000800c: ('CALG_SHA_256', hashlib.sha256),
            0x0000800d: ('CALG_SHA_384', hashlib.sha384),
            0x0000800e: ('CALG_SHA_512', hashlib.sha512),
            0x00008003: ('CALG_MD5', hashlib.md5)
        }

        hProv, Algid, hKey, dwFlags, phHash = argv
        argv[1] = hash_algs.get(Algid, Algid)[0]

        if hKey != 0:
            return 0

        if Algid not in hash_algs:
            emu.set_last_error(adv32.NTE_BAD_ALGID)
            return 0

        hnd = self.get_handle()
        self.hash_objects.update({hnd: hash_algs[Algid][1]()})
        self.mem_write(phHash, hnd.to_bytes(self.get_ptr_size(), 'little'))
        return 1

    @apihook('CryptHashData', argc=4)
    def CryptHashData(self, emu, argv, ctx={}):
        '''
        BOOL CryptHashData(
          HCRYPTHASH hHash,
          const BYTE *pbData,
          DWORD      dwDataLen,
          DWORD      dwFlags
        );
        '''

        hHash, pbData, dwDataLen, dwFlags = argv
        hnd = self.hash_objects.get(hHash, None)
        if hnd is None:
            emu.set_last_error(windefs.ERROR_INVALID_HANDLE)
            return 0

        if dwDataLen <= 0:
            return 0

        data = self.mem_read(pbData, dwDataLen)
        hnd.update(data)
        return 1

    @apihook('CryptGetHashParam', argc=5)
    def CryptGetHashParam(self, emu, argv, ctx={}):
        '''
        BOOL CryptGetHashParam(
          HCRYPTHASH hHash,
          DWORD      dwParam,
          BYTE       *pbData,
          DWORD      *pdwDataLen,
          DWORD      dwFlags
        );
        '''
        hHash, dwParam, pbData, pdwDataLen, dwFlags = argv

        param_enums = {
            1: "HP_ALGID",
            2: "HP_HASHVAL",
            4: "HP_HASHSIZE",
            5: "HP_HMAC_INFO"
        }

        if dwParam in param_enums.keys():
            argv[1] = param_enums[dwParam]

        return 1

    @apihook('CryptDestroyHash', argc=1)
    def CryptDestroyHash(self, emu, argv, ctx={}):
        """
        BOOL CryptDestroyHash(
          HCRYPTHASH hHash
        );
        """
        hHash = argv

        return 1

    @apihook('CryptDeriveKey', argc=5)
    def CryptDeriveKey(self, emu, argv, ctx={}):
        """
        BOOL CryptDeriveKey(
          HCRYPTPROV hProv,
          ALG_ID     Algid,
          HCRYPTHASH hBaseData,
          DWORD      dwFlags,
          HCRYPTKEY  *phKey
        );
        """

        hProv, Algid, hBaseData, dwFlags, phKey = argv

        # Only RC4 supported right now
        if Algid != 0x6801:
            return 0

        hnd = self.hash_objects.get(hBaseData, None)

        if hnd is None:
            emu.set_last_error(windefs.ERROR_INVALID_HANDLE)
            return 0

        # CryptDeriveKey zeroes out the last 11 bytes of the hash,
        # so we gotta do the same before it is written to the
        # phKey structure
        fixed_digest = hnd.digest()[:5] + b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'

        ptrsz = emu.get_ptr_size()

        hKey = self.win.HCRYPTKEY(ptrsz)
        hKey.Algid = Algid
        hKey.keylen = hnd.digest_size
        hKey.keyp = self.mem_alloc(hKey.keylen)

        hKeyp = self.mem_alloc(hKey.sizeof())

        self.mem_write(hKey.keyp, fixed_digest)

        self.mem_write(hKeyp, hKey.get_bytes())
        self.mem_write(phKey, hKeyp.to_bytes(ptrsz, "little"))

        return 1

    @apihook('CryptDecrypt', argc=6)
    def CryptDecrypt(self, emu, argv, ctx={}):
        """
        BOOL CryptDecrypt(
          HCRYPTKEY  hKey,
          HCRYPTHASH hHash,
          BOOL       Final,
          DWORD      dwFlags,
          BYTE       *pbData,
          DWORD      *pdwDataLen
        );
        """

        hKey, hHash, Final, dwFlags, pbData, pdwDataLen = argv

        # Hashing not supported
        if hHash:
            return 0

        ptrsz = emu.get_ptr_size()

        hKey = self.mem_cast(self.win.HCRYPTKEY(ptrsz), hKey)

        # Only RC4 supported right now
        if hKey.Algid != 0x6801:
            return 0

        encdatalen_b = self.mem_read(pdwDataLen, 4)
        encdatalen = int.from_bytes(encdatalen_b, "little")

        encdata = self.mem_read(pbData, encdatalen)

        key = self.mem_read(hKey.keyp, hKey.keylen)

        if self.rc4 is None:
            self.rc4 = ARC4.new(key)

        dec = self.rc4.decrypt(encdata)
        declen = len(dec)

        self.mem_write(pbData, dec)
        self.mem_write(pdwDataLen, int.to_bytes(declen, 4, "little"))

        if Final == True:
            self.rc4 = None

        return 1

    @apihook('RegGetValue', argc=7, conv=_arch.CALL_CONV_STDCALL)
    def RegGetValue(self, emu, argv, ctx={}):
        '''
        LSTATUS RegGetValueW(
            HKEY    hkey,
            LPCWSTR lpSubKey,
            LPCWSTR lpValue,
            DWORD   dwFlags,
            LPDWORD pdwType,
            PVOID   pvData,
            LPDWORD pcbData
            );
        '''

        hKey, lpSubKey, lpValue, dwFlags, lpType, lpData, lpcbData = argv
        rv = windefs.ERROR_SUCCESS

        cw = self.get_char_width(ctx)
        if lpSubKey:
            lpSubKey = self.read_mem_string(lpSubKey, cw)
            argv[1] = lpSubKey

        if lpValue:
            lpValue = self.read_mem_string(lpValue, cw)
            argv[2] = lpValue

        type_name = regdefs.get_value_type(lpType)
        if type_name:
            argv[4] = type_name

        length = 0
        if lpcbData:
            length = self.mem_read(lpcbData, 4)
            length = int.from_bytes(length, 'little')

        key = self.reg_get_key(hKey)
        if key:
            val = key.get_value(lpValue)
            if val:
                output = b''

                if lpcbData:
                    self.mem_write(lpcbData, len(output).to_bytes(4, 'little'))

                if len(output) > length:
                    rv = windefs.ERROR_INSUFFICIENT_BUFFER
                else:
                    self.mem_write(lpData, output)

            # For now, return an empty buffer
            else:
                output = b'\x00' * length
                self.mem_write(lpData, output)
                rv = windefs.ERROR_SUCCESS

            kp = key.get_path()
            self.log_registry_access(kp, 'read_value', value_name=lpValue, size=length,
                                     buffer=lpData)

        return rv

    @apihook('EnumServicesStatus', argc=8, conv=_arch.CALL_CONV_STDCALL)
    def EnumServicesStatus(self, emu, argv, ctx={}):
        '''
        BOOL EnumServicesStatusA(
          SC_HANDLE              hSCManager,
          DWORD                  dwServiceType,
          DWORD                  dwServiceState,
          LPENUM_SERVICE_STATUSA lpServices,
          DWORD                  cbBufSize,
          LPDWORD                pcbBytesNeeded,
          LPDWORD                lpServicesReturned,
          LPDWORD                lpResumeHandle
        );
        '''
        hSCManager, dwServiceType, dwServiceState, lpServices, cbBufSize, \
            pcbBytesNeeded, lpServicesReturned, lpResumeHandle = argv

        service_type_str = adv32.get_define_int(dwServiceType, 'SERVICE_')
        if service_type_str:
            argv[1] = service_type_str

        service_state_str = adv32.get_define_int(dwServiceState, 'SERVICE_')
        if service_state_str:
            argv[2] = service_state_str

        # TODO: Populate service status output
        return 1

    @apihook('OpenService', argc=3, conv=_arch.CALL_CONV_STDCALL)
    def OpenService(self, emu, argv, ctx={}):
        '''
        SC_HANDLE OpenServiceA(
          SC_HANDLE hSCManager,
          LPCSTR    lpServiceName,
          DWORD     dwDesiredAccess
        );
        '''
        hSCManager, lpServiceName, dwDesiredAccess = argv
        cw = self.get_char_width(ctx)
        svcname = self.read_mem_string(lpServiceName, cw)
        argv[1] = svcname
        return self.get_handle()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from itertools import groupby
from operator import itemgetter

import speakeasy.common as common


class MemMap(object):
    """
    Class that defines a memory mapping (e.g. heap/pool alloc, binary image, etc.)
    """
    def __init__(self, base, size, tag, prot, flags, block_base, block_size,
                 shared=False, process=None):
        self.base = base
        self.size = size

        base_addr_tag = '.0x%x' % (base)
        if tag and base_addr_tag not in tag:
            tag += base_addr_tag

        if tag:
            tag = list(tag)
            bad_chars = '\\?[]:]'
            [tag.__setitem__(j, '_') for j in [i for i, e in enumerate(tag) if e in bad_chars]]
            tag = ''.join(tag)

        self.tag = tag
        self.prot = prot
        self.flags = flags
        self.shared = shared
        self.free = False
        self.process = process
        self.block_base = block_base
        self.block_size = block_size

    def update_tag(self, new_tag):
        """
        Set the tag for the memory mapping
        """
        self.tag = new_tag

    def get_process(self):
        """
        Get the process object associated with a memory map
        """
        return self.process

    def set_process(self, process):
        """
        Set the process object associated with a memory map
        """
        self.process = process

    def get_tag(self):
        """
        Get the tag for the memory mapping
        """
        return self.tag

    def get_prot(self):
        """
        Get the memory permissions for a map
        """
        return self.prot

    def get_flags(self):
        """
        Get the memory flags for a map
        """
        return self.flags

    def get_size(self):
        """
        Get the byte size for the current memory mapping
        """
        return self.size

    def get_base(self):
        """
        Get the base address (lowest possible address) of the current memory map
        """
        return self.base

    def set_alloc(self):
        """
        Set the current mapping to be in an allocated state
        """
        self.free = False

    def set_free(self):
        """
        Set the current mapping to be in a free state
        """
        self.free = True

    def is_free(self):
        """
        Return the alloc state of a memory block
        """
        return self.free

    def __hash__(self):
        return hash(self.base)

    def __eq__(self, other):
        if other is not None:
            return self.base == other.base

    def __ne__(self, other):
        return not(self == other)


class MemoryManager(object):

    """
    Primitive memory manager used to block OS sized allocation units into something more practical
    """

    def __init__(self, *args, **kwargs):
        super(MemoryManager, self).__init__()
        self.maps = []
        self.emu_eng = None
        self.mem_reserves = []
        self.block_base = 0
        self.block_size = 0
        self.block_offset = 0
        self.page_size = 0x1000

    def _hook_mem_map_dispatch(self, mm):
        hl = self.hooks.get(common.HOOK_MEM_MAP, [])
        ctx = {}
        for mem_map_hook in hl:
            if mem_map_hook.enabled:
                # the mapped memory region's base address falls within the hook's bounds
                if mem_map_hook.begin <= mm.get_base():
                    if not mem_map_hook.end or mem_map_hook.end > mm.get_base():
                        mem_map_hook.cb(self, mm.get_base(), mm.get_size(),
                                        mm.get_tag(), mm.get_prot(), mm.get_flags(), ctx)

    def mem_map(self, size, base=None, perms=common.PERM_MEM_RWX,
                tag=None, flags=0, shared=False, process=None):
        """
        Map a block of memory with specified permissions and a tag
        """
        if not process and tag and not tag.startswith('emu'):
            process = self.get_current_process()

        if base is None:
            if size < self.page_size and size % self.page_size:
                addr = self.block_base + self.block_offset
                pad_size = 0x10 - (size % 0x10)
                size += pad_size
                if not self.block_base or ((addr + size) > self.block_base + self.page_size):
                    block = self.get_valid_ranges(self.page_size)
                    self.block_base, self.block_size = block

                    self.emu_eng.mem_map(self.block_base, self.block_size)
                    self.block_offset = 0
                    addr = self.block_base + self.block_offset

                self.block_offset += size
                base = addr

                mm = MemMap(base, size, tag, perms, flags, self.block_base, self.block_size,
                            shared, process)

                self.maps.append(mm)
                self._hook_mem_map_dispatch(mm)
                return base

        block = self.get_valid_ranges(size, addr=base)
        base, size = block

        block_size = self.block_size
        if size > self.block_size:
            block_size = size
        mm = MemMap(base, size, tag, perms, flags, base, block_size, shared, process)
        self.emu_eng.mem_map(base, size, perms=perms)
        self.maps.append(mm)
        self._hook_mem_map_dispatch(mm)
        return base

    def mem_free(self, base):
        """
        Free a block of memory, if all blocks in a block are set to free, unmap the entire block
        """
        mm = self.get_address_map(base)
        if mm:
            mm.set_free()

            # If we want to freeze memory, just return
            if self.keep_memory_on_free:
                return

            ml = [m for m in self.get_mem_maps() if m.block_base == mm.block_base]
            # if all blocks are free in the current block, free it from the emu engine
            if all([m.free for m in ml]):
                self.block_base = 0
                self.mem_unmap(mm.block_base, mm.block_size)
                [self.maps.remove(mm) for mm in ml]

    def mem_remap(self, frm, to):
        """
        Remap a block of emulated memory, and return the new address,
        or -1 on error
        Protections remain the same
        """
        map = self.get_address_map(frm)

        if not map:
            return -1

        prot = map.prot
        size = map.size

        # Exclude old memory region in tag name
        tag = map.tag[:map.tag.rfind(".")]

        contents = self.mem_read(map.base, size)

        # Will unmap as well
        self.mem_free(map.base)

        newmem = self.mem_map(size, base=to, perms=prot, tag=tag)
        
        if newmem != to:
            return -1

        self.mem_write(newmem, contents)

        return newmem

    def mem_unmap(self, base, size):
        """
        Free a block of emulated memory
        """
        self.emu_eng.mem_unmap(base, size)

    def mem_write(self, addr, data):
        """
        Write bytes into the emulated address space
        """
        self.emu_eng.mem_write(addr, data)

    def mem_read(self, addr, size):
        """
        Read bytes from the emulated address space
        """
        return bytes(self.emu_eng.mem_read(addr, size))

    def mem_protect(self, addr, size, perms):
        """
        Change memory protections
        """
        self.emu_eng.mem_protect(addr, size, perms)

    def _mem_unmap_region(self, base, size):
        """
        Remove an entire memory region that may not have blocks allocated within it
        """
        self.emu_eng.mem_unmap(base, size)

    def get_address_map(self, address):
        """
        Get the "MemMap" object associated with a specific address
        """
        for m in self.maps:
            if m.base <= address <= (m.base + m.size) - 1:
                return m

    def get_reserve_map(self, address):
        """
        Get the "MemMap" object that was only reserved for a specific address
        """
        for m in self.mem_reserves:
            if m.base <= address <= (m.base + m.size) - 1:
                return m

    def is_address_valid(self, address):
        """
        Was this address previously reserved or mapped?
        """
        if self.get_address_map(address):
            return True
        if self.get_reserve_map(address):
            return True
        return False

    def get_address_tag(self, address):
        """
        Get the tag for a supplied memory address
        """
        for m in self.maps:
            if address >= m.base and address <= (m.base + m.size) - 1:
                return m.tag

    def mem_reserve(self, size, base=None, perms=None, tag=None, flags=0, shared=False):
        """
        Reserve (but do not map) a block of memory
        """
        if base is None:
            block = self.get_valid_ranges(size)
            base, size = block

        mm = MemMap(base, size, tag, perms, flags, base, self.block_size, shared)

        self.mem_reserves.append(mm)
        return base

    def purge_memory(self):
        """
        Unmap all current blocks of mapped memory
        """
        for region in self.get_mem_regions():
            base, end, perms = region
            size = (end - base) + 1
            self._mem_unmap_region(base, size)

    def get_mem_maps(self):
        """
        Get the listing of current memory maps
        """
        return self.maps

    def mem_map_reserve(self, mapped_base):
        """
        Map a previously reserved block of memory
        """
        for r in self.mem_reserves:
            if mapped_base == r.base:
                self.mem_reserves.remove(r)
                return self.mem_map(r.size, base=r.base, perms=r.prot, tag=r.tag)
        return None

    def get_mem_regions(self):
        """
        Get the current regions of mapped memory
        """
        return self.emu_eng.mem_regions()

    def get_valid_ranges(self, size, addr=None):
        """
        Retrieve a valid address range that can satisfy the requested size.
        Optionally, a base address can be specified to test if it can be used
        """

        def get_runs(i):
            for k, g in groupby(enumerate(i),
                                lambda ix: ix[0] - (ix[1] >> 12)):
                yield tuple(map(itemgetter(1), g))

        page_size = self.page_size

        # mem_map needs to be page aligned
        total = size

        # alloced address needs to also be on a page boundary
        if addr is None:
            addr = page_size
        base = addr - (addr % page_size)

        if total < page_size:
            total = page_size
        elif total % page_size:
            total += (page_size - (total % page_size))

        curr = []
        for m in self.get_mem_regions():
            curr += range(m[0], m[1], page_size)

        # Add reserved memory so we don't accidentally allocate it
        for res in self.mem_reserves:
            curr += range(res.base, (res.base + res.size), page_size)

        curr = sorted(set(curr))

        attempts = 9999
        while attempts:
            req = set(range(base, base + total, page_size))

            diffs = sorted(req.difference(curr))

            if len(diffs) == len(req):
                break

            if not attempts % 10:
                base += (page_size * 1000)
            else:
                base += total
            attempts -= 1

        if attempts == 0:
            raise Exception('Failed to allocate emulator memory')

        a = [r for r in get_runs(diffs)][0]

        return (min(a), total)
